"""Module providing functionality to interact with IBM Quantum and IBM Cloud devices.

Note:
    This module requires additional dependencies from ``ket-lang[ibm]``.

    Install with: ``pip install ket-lang[ibm]``.

"""

from __future__ import annotations

# SPDX-FileCopyrightText: 2024 Evandro Chagas Ribeiro da Rosa <evandro@quantuloop.com>
# SPDX-FileCopyrightText: 2024 Otávio Augusto de Santana Jatobá
# <otavio.jatoba@grad.ufsc.br>
#
# SPDX-License-Identifier: Apache-2.0

try:
    from qiskit import QuantumCircuit
    from qiskit.providers import Backend
    from qiskit.circuit import library
except ImportError as exc:
    raise ImportError(
        "IBMDevice requires the qiskit module to be used. You can install them"
        "alongside ket by running `pip install ket[ibm]`."
    ) from exc

import json
from ctypes import CFUNCTYPE, POINTER, c_uint8, c_size_t, c_double
from ..clib.libket import BatchCExecution, make_configuration, BatchExecution
from .ibm_client import IBMClient

__all__ = ["IBMDevice"]


class IBMDevice:  # pylint: disable=too-many-instance-attributes
    """IBM Qiskit backend for Ket process.

    Args:
        backend: The backend to be used for the quantum execution.
        num_qubits: The number of qubits to be used for the quantum circuit.
        use_qiskit_transpiler: Use Qiskit transpiler instead of Ket's.
    """

    def __init__(
        self,
        backend: Backend,
        num_qubits: int | None = None,
        *,
        use_qiskit_transpiler: bool = False,
    ) -> None:
        self.num_qubits = (
            num_qubits if num_qubits is not None else backend.configuration().n_qubits
        )
        self.client = None
        self._backend = backend

        if not use_qiskit_transpiler:
            self.coupling_graph = (
                list(backend.coupling_map.graph.edge_list())
                if backend.coupling_map
                else [
                    [i, j]
                    for i in range(self.num_qubits)
                    for j in range(self.num_qubits)
                    if i != j
                ]
            )
        else:
            self.coupling_graph = None

        self._result_json = None
        self._result_json_len = None
        self._formatted_result = None

        @CFUNCTYPE(
            None,
            POINTER(c_uint8),
            c_size_t,
            POINTER(c_uint8),
            c_size_t,
            POINTER(c_double),
            c_size_t,
        )
        def submit_execution(
            logical_circuit,
            logical_circuit_size,
            physical_circuit,
            physical_circuit_size,
            _parameters,
            _parameters_size,
        ):
            """Sends the ket circuit instructions from libket to the IBM Client."""
            logical_circuit = json.loads(
                bytearray(logical_circuit[:logical_circuit_size])
            )
            physical_circuit = json.loads(
                bytearray(physical_circuit[:physical_circuit_size])
            )
            self.client = IBMClient(self.num_qubits, self._backend)
            self._formatted_result = self.client.process_instructions(
                physical_circuit if physical_circuit is not None else logical_circuit
            )

        @CFUNCTYPE(
            None,
            POINTER(POINTER(c_uint8)),
            POINTER(c_size_t),
        )
        def get_result(result_ptr, size):
            """Sends the processing result from the IBM Client back to libket."""
            result_dict = self._formatted_result
            result_json = json.dumps(result_dict).encode("utf-8")
            result_len = len(result_json)

            self._result_json = (c_uint8 * result_len)(*result_json)
            self._result_json_len = result_len
            # Set the result pointer and size, both of which must remain valid inside
            # python until the libket process has finished. That's why they're class
            # attributes.
            result_ptr[0] = self._result_json
            size[0] = self._result_json_len

        @CFUNCTYPE(None)
        def clear():
            pass

        self.c_struct = BatchCExecution(
            submit_execution,
            get_result,
            clear,
        )

    def configure(self):
        """Set up Ket process.

        Example:

            .. code-block:: python

                device = IBMDevice(backend)
                process = Process(device.configure())


        """

        return make_configuration(
            num_qubits=self.num_qubits,
            batch_execution=self.c_struct,
            measure="Allowed",
            sample="Allowed",
            exp_value="Allowed",
            dump="Disable",
            define_qpu=self.coupling_graph is not None,
            coupling_graph=self.coupling_graph,
            u4_gate_type="CX",
            u2_gate_set="All",
            gradient="Disable",
        )

    @property
    def circuit(self) -> QuantumCircuit:
        """Quantum circuit object for the IBM device."""
        return self.client.circuit

    @property
    def isa_circuit(self) -> QuantumCircuit:
        """Quantum circuit object for the IBM device."""
        return self.client.isa_circuit

    @property
    def backend(self) -> Backend:
        """Backend object for the IBM device."""
        return self._backend


class IBMDeviceForDraw(BatchExecution):
    """IBMDevice use for qulib.draw"""

    def __init__(
        self,
        num_qubits: int,
        decompose: bool = False,
    ):
        super().__init__()
        self.circuit = QuantumCircuit(num_qubits)
        self.num_qubits = num_qubits
        self.decompose = decompose

    def clear(self):
        pass

    def submit_execution(self, logical_circuit, _fc, _p):
        self.process_instructions(logical_circuit)

    def get_result(self):
        return {
            "measurements": [],
            "exp_values": [],
            "samples": [],
            "dumps": [],
            "gradients": None,
        }

    def pauli_x(self, target, control):
        gate = library.XGate()
        if len(control) >= 1:
            gate = gate.control(len(control))
        self.circuit.append(gate, control + [target])

    def pauli_y(self, target, control):
        gate = library.YGate()
        if len(control) >= 1:
            gate = gate.control(len(control))
        self.circuit.append(gate, control + [target])

    def pauli_z(self, target, control):
        gate = library.ZGate()
        if len(control) >= 1:
            gate = gate.control(len(control))
        self.circuit.append(gate, control + [target])

    def hadamard(self, target, control):
        gate = library.HGate()
        if len(control) >= 1:
            gate = gate.control(len(control))
        self.circuit.append(gate, control + [target])

    def rotation_x(self, target, control, **kwargs):
        gate = library.RXGate(kwargs["Value"])
        if len(control) >= 1:
            gate = gate.control(len(control))
        self.circuit.append(gate, control + [target])

    def rotation_y(self, target, control, **kwargs):
        gate = library.RYGate(kwargs["Value"])
        if len(control) >= 1:
            gate = gate.control(len(control))
        self.circuit.append(gate, control + [target])

    def rotation_z(self, target, control, **kwargs):
        gate = library.RZGate(kwargs["Value"])
        if len(control) >= 1:
            gate = gate.control(len(control))
        self.circuit.append(gate, control + [target])

    def phase(self, target, control, **kwargs):
        gate = library.PhaseGate(kwargs["Value"])
        if len(control) >= 1:
            gate = gate.control(len(control))
        self.circuit.append(gate, control + [target])

    def config(
        self,
    ):
        """Configure process"""
        return super().configure(
            num_qubits=self.num_qubits,
            measure="Disable",
            sample="Disable",
            exp_value="Disable",
            dump="Disable",
            gradient="Disable",
            define_qpu=self.decompose,
            coupling_graph=None,
            u4_gate_type="CX",
            u2_gate_set="All",
        )



# SPDX-FileCopyrightText: 2024 Evandro Chagas Ribeiro da Rosa <evandro@quantuloop.com>
#
# SPDX-License-Identifier: Apache-2.0

from math import sqrt, pi, cos, sin
from cmath import exp, isclose
import ket
from ket import qulib

GATES = {
    ket.X: [[0, 1], [1, 0]],
    ket.Y: [[0, -1j], [1j, 0]],
    ket.Z: [[1, 0], [0, -1]],
    ket.H: [[1 / sqrt(2), 1 / sqrt(2)], [1 / sqrt(2), -1 / sqrt(2)]],
    ket.T: [[1, 0], [0, exp(1j * pi / 4)]],
    ket.S: [[1, 0], [0, 1j]],
}

ROTATION_GATES = {
    ket.RX: lambda theta: [
        [cos(theta / 2), -1j * sin(theta / 2)],
        [-1j * sin(theta / 2), cos(theta / 2)],
    ],
    ket.RY: lambda theta: [
        [cos(theta / 2), -sin(theta / 2)],
        [sin(theta / 2), cos(theta / 2)],
    ],
    ket.RZ: lambda theta: [[exp(-1j * theta / 2), 0], [0, exp(1j * theta / 2)]],
}


def linspace(start, stop, num):
    yield start
    step = (stop - start) / (num - 1)
    next_num = start
    for _ in range(num - 1):
        next_num += step
        yield next_num


def test_decomposition_su2():
    n = 7

    for gate, mat_gate in ROTATION_GATES.items():
        for ang in linspace(0.0, 2 * pi, 8):
            matrix = mat_gate(ang)

            ctrl_gate = lambda q: ket.ctrl(q[:-1], gate(ang))(q[-1])
            result_matrix = ket.qulib.dump_matrix(ctrl_gate, num_qubits=n)

            gate_matrix = [
                [result_matrix[-2][-2], result_matrix[-2][-1]],
                [result_matrix[-1][-2], result_matrix[-1][-1]],
            ]

            eye_matrix = result_matrix

            eye_matrix[-2][-2] = 1
            eye_matrix[-2][-1] = 0
            eye_matrix[-1][-2] = 0
            eye_matrix[-1][-1] = 1

            eye = list(
                list(1 if i == j else 0 for i in range(2**n)) for j in range(2**n)
            )

            assert all(
                isclose(gate_matrix[i][j], matrix[i][j], abs_tol=1e-10)
                for i in range(2)
                for j in range(2)
            )

            assert all(
                isclose(eye_matrix[i][j], eye[i][j], abs_tol=1e-10)
                for i in range(2**n)
                for j in range(2**n)
            )


def test_decomposition_c_t():

    n = 7

    for ket_gate, matrix in GATES.items():
        gate = lambda q: ket.ctrl(q[:-1], ket_gate)(q[-1])
        result_matrix = ket.qulib.dump_matrix(gate, num_qubits=n)

        gate_matrix = [
            [result_matrix[-2][-2], result_matrix[-2][-1]],
            [result_matrix[-1][-2], result_matrix[-1][-1]],
        ]

        eye_matrix = result_matrix

        eye_matrix[-2][-2] = 1
        eye_matrix[-2][-1] = 0
        eye_matrix[-1][-2] = 0
        eye_matrix[-1][-1] = 1

        eye = list(list(1 if i == j else 0 for i in range(2**n)) for j in range(2**n))

        assert all(
            isclose(gate_matrix[i][j], matrix[i][j], abs_tol=1e-10)
            for i in range(2)
            for j in range(2)
        )

        assert all(
            isclose(eye_matrix[i][j], eye[i][j], abs_tol=1e-10)
            for i in range(2**n)
            for j in range(2**n)
        )


def test_decomposition_t_c():
    n = 7

    for ket_gate in GATES.keys():
        gate = lambda q: ket.ctrl(q[1:], ket_gate)(q[0])

        decompose_matrix = qulib.dump_matrix(gate, num_qubits=n)

        not_decompose_matrix = qulib.dump_matrix(gate, num_qubits=n)

        assert all(
            isclose(decompose_matrix[i][j], not_decompose_matrix[i][j], abs_tol=1e-10)
            for i in range(2**n)
            for j in range(2**n)
        )


if __name__ == "__main__":
    test_decomposition_c_t()
    # test_decomposition_t_c()
    # test_decomposition_is_enabled()
    # test_decomposition_su2()
    print("Ok")



"""Builder for qiskit quantum circuits."""

from __future__ import annotations

# SPDX-FileCopyrightText: 2024 Evandro Chagas Ribeiro da Rosa <evandro@quantuloop.com>
# SPDX-FileCopyrightText: 2024 Otávio Augusto de Santana Jatobá
# <otavio.jatoba@grad.ufsc.br>
#
# SPDX-License-Identifier: Apache-2.0

try:
    from qiskit import QuantumCircuit
    from qiskit.circuit import library, Gate
    from qiskit.quantum_info import SparsePauliOp
except ImportError as exc:
    raise ImportError(
        "QiskitBuilder requires the qiskit module to be used. You can install them"
        "alongside ket by running `pip install ket[ibm]`."
    ) from exc
from typing import Any


class QiskitBuilder:
    """Builder for qiskit quantum circuits from ket-lang instructions."""

    def __init__(self, num_qubits: int) -> None:
        self.num_qubits = num_qubits
        self.circuit = QuantumCircuit(num_qubits, num_qubits)

    def build(self, instructions, meas_map: dict, sample_map: dict) -> dict[str, Any]:
        """Builds the Qiskit quantum circuit from the ket-lang instructions."""

        data = {
            "circuit": self.circuit,
            "observables": [],
        }

        for inst in instructions:
            if "Identity" in inst:
                continue

            if "Gate" in inst:
                gate_type = inst["Gate"]["gate"]
                gate: Gate = self.get_gate(gate_type)

                control = inst["Gate"]["control"]
                if len(control):
                    gate = gate.control(len(control))
                control = [self._get_qubit_index(qubit) for qubit in control]
                data["circuit"].append(
                    gate, control + [self._get_qubit_index(inst["Gate"]["target"])]
                )

            elif "Measure" in inst:
                qubits = [
                    self._get_qubit_index(qubit) for qubit in inst["Measure"]["qubits"]
                ]
                meas_map[inst["Measure"]["index"]] = qubits
                data["circuit"].measure(qubits, qubits)

            elif "ExpValue" in inst:
                hamiltonian: dict = inst["ExpValue"]["hamiltonian"]
                data["observables"].append(self.build_observable(hamiltonian))

            elif "Sample" in inst:
                qubits = [
                    self._get_qubit_index(qubit) for qubit in inst["Sample"]["qubits"]
                ]
                sample_map[inst["Sample"]["index"]] = qubits
                sample_map["shots"] = max(
                    sample_map.get("shots", 2048), inst["Sample"]["shots"]
                )
                data["circuit"].measure(qubits, qubits)

            elif "Dump" in inst:
                raise RuntimeError("Operation not supported")

            else:
                raise RuntimeError("Unknown operation")

        return data

    @staticmethod
    def _get_qubit_index(qubit):
        try:
            return qubit["index"]
        except KeyError:
            return qubit["Main"]["index"]

    def get_gate(self, gate_type) -> Gate:
        """From the type of the gate, returns the corresponding qiskit gate object."""

        if "PauliX" in gate_type:
            return library.XGate()
        if "PauliY" in gate_type:
            return library.YGate()
        if "PauliZ" in gate_type:
            return library.ZGate()
        if "Hadamard" in gate_type:
            return library.HGate()
        if isinstance(gate_type, dict) and any(
            "Rotation" in key or "Phase" in key for key in list(gate_type.keys())
        ):
            angle = list(gate_type.values())[0]["Value"]
            return self._get_rotation_or_phase_gate(gate_type, angle)

        raise RuntimeError("Unknown gate")

    def _get_rotation_or_phase_gate(self, gate_type, theta) -> Gate:
        """From the type of the gate, returns the corresponding qiskit rotation or
        phase gate object."""

        if "RotationX" in gate_type:
            return library.RXGate(theta)
        if "RotationY" in gate_type:
            return library.RYGate(theta)
        if "RotationZ" in gate_type:
            return library.RZGate(theta)

        return library.U1Gate(theta)  # Phase gate

    def build_observable(self, hamiltonian: dict[str, Any]) -> SparsePauliOp:
        """Builds a Qiskit compliant observable format from the ket-lang hamiltonian
        format.
        ---
        Details:
        -
        The ket hamiltonian format is a dictionary with the following keys:
        - "coefficients": List of coefficients for each product term.
        - "products": List of product terms, where each term is a list of dictionaries
          with the following keys:
            - "qubit": The qubit index.
            - "pauli": The Pauli operator acting on the qubit.

        The qiskit observable is a SparsePauliOp object, made upon a list of Pauli
        strings and a list of coefficients. The Pauli strings are made by concatenating
        the Pauli operators acting on each qubit in the system. The coefficients are
        spread across the Pauli strings, with each corresponding coefficient repeated
        for each Pauli string in the product term.

        Example:
        -
        Given the following ket hamiltonian in a system with 3 qubits:

        hamiltonian = {
            "coefficients": [1, 2],
            "products": [
                [{"qubit": 0, "pauli": "PauliX"}
                {"qubit": 1, "pauli": "PauliX"},
                {"qubit": 1, "pauli": "PauliZ"}],

                [{"qubit": 0, "pauli": "PauliY"}
                {"qubit": 1, "pauli": "PauliY"},
                {"qubit": 2, "pauli": "PauliZ"}],
            ],
        }

        The qiskit observable will be build upon the following components:
        - qiskit_hamiltonian = ["XXI", "IZI, "YYZ"]
        - qiskit_coeffs = [1, 1, 2]

        The qiskit_hamiltonian and qiskit_coeffs are used to create a SparsePauliOp
        object, which is then used to calculate the expectation value.
        """
        hamiltonian_sum: list[list[dict]] = hamiltonian["products"]
        qiskit_hamiltonian = []
        qiskit_coeffs = []

        for coef_index, product in enumerate(hamiltonian_sum):
            qiskit_hmlt_product = []
            i = 0
            while i < len(product):
                qiskit_hamlt_term = ["I"] * self.num_qubits
                while i < len(product):
                    current_qubit: int = self._get_qubit_index(product[i]["qubit"])
                    if qiskit_hamlt_term[current_qubit] != "I":
                        break
                    pauli_type: str = product[i]["pauli"][-1]
                    qiskit_hamlt_term[current_qubit] = pauli_type
                    i += 1
                qiskit_hmlt_product.append("".join(reversed(qiskit_hamlt_term)))

            qiskit_hamiltonian.extend(qiskit_hmlt_product)
            qiskit_coeffs.extend(
                [hamiltonian["coefficients"][coef_index]] * len(qiskit_hmlt_product)
            )

        observable = SparsePauliOp(qiskit_hamiltonian, qiskit_coeffs)
        return observable



"""Wrapper for Libket C API."""

from __future__ import annotations

# SPDX-FileCopyrightText: 2020 Evandro Chagas Ribeiro da Rosa <evandro@quantuloop.com>
# SPDX-FileCopyrightText: 2020 Rafael de Santiago <r.santiago@ufsc.br>
#
# SPDX-License-Identifier: Apache-2.0

from ctypes import (
    CFUNCTYPE,
    Structure,
    c_uint32,
    c_void_p,
    c_size_t,
    POINTER,
    c_bool,
    c_uint8,
    c_int32,
    c_uint64,
    c_double,
)
from functools import reduce
import json
from operator import iconcat
from typing import Literal
import weakref
from os import environ
from os.path import dirname
from abc import ABC, abstractmethod

from .wrapper import load_lib, os_lib_name


HADAMARD = 0
PAULI_X = 1
PAULI_Y = 2
PAULI_Z = 3
ROTATION_X = 10
ROTATION_Y = 20
ROTATION_Z = 30
PHASE_SHIFT = 31

SUCCESS = 0
CONTROL_TWICE = 1
DATA_NOT_AVAILABLE = 2
DEALLOCATED_QUBIT = 3
QUBIT_INDEX_OUT_OF_BOUNDS = 4
NUMBER_OF_QUBITS_EXCEEDED = 5
NO_ADJ = 6
NO_CTRL = 7
NON_GATE_INSTRUCTION_IN_ADJ = 8
TARGET_IN_CONTROL = 9
PROCESS_READY_TO_EXECUTE = 10
UNEXPECTED_RESULT_DATA = 11
DUMP_NOT_ALLOWED = 12
EXP_VALUE_NOT_ALLOWED = 13
SAMPLE_NOT_ALLOWED = 14
MEASURE_NOT_ALLOWED = 15
UNDEFINED_ERROR = 16


class BatchCExecution(Structure):  # pylint: disable=too-few-public-methods
    """C BatchCExecution Structure"""

    _fields_ = [
        (
            "submit_execution",
            CFUNCTYPE(
                None,
                POINTER(c_uint8),
                c_size_t,
                POINTER(c_uint8),
                c_size_t,
                POINTER(c_double),
                c_size_t,
            ),
        ),
        ("get_results", CFUNCTYPE(None, POINTER(POINTER(c_uint8)), POINTER(c_size_t))),
        ("clear", CFUNCTYPE(None)),
    ]


API_argtypes = {
    # 'ket_type_method': ([input_list], [output_list]),
    "ket_set_log_level": ([c_uint32], []),
    "ket_build_info": ([], [POINTER(c_uint8), c_size_t]),
    "ket_process_new": ([c_void_p], [c_void_p]),
    "ket_process_delete": ([c_void_p], []),
    "ket_process_allocate_qubit": ([c_void_p], [c_size_t]),
    "ket_process_apply_gate": (
        [
            c_void_p,  # process
            c_int32,  # gate
            c_double,  # angle
            c_bool,  # use_param
            c_size_t,  # param_index
            c_size_t,  # target
        ],
        [],
    ),
    "ket_process_set_parameter": ([c_void_p, c_double], [c_size_t]),
    "ket_process_get_gradient": ([c_void_p, c_size_t], [c_bool, c_double]),
    "ket_process_apply_global_phase": (
        [c_void_p, c_double],
        [],
    ),
    "ket_process_measure": ([c_void_p, POINTER(c_size_t), c_size_t], [c_size_t]),
    "ket_hamiltonian_new": ([], [c_void_p]),
    "ket_hamiltonian_add": (
        [c_void_p, POINTER(c_int32), c_size_t, POINTER(c_size_t), c_size_t, c_double],
        [],
    ),
    "ket_process_exp_value": (
        [c_void_p, c_void_p],
        [c_size_t],
    ),
    "ket_process_sample": (
        [c_void_p, POINTER(c_size_t), c_size_t, c_uint64],
        [c_size_t],
    ),
    "ket_process_dump": ([c_void_p, POINTER(c_size_t), c_size_t], [c_size_t]),
    "ket_process_ctrl_push": ([c_void_p, POINTER(c_size_t), c_size_t], []),
    "ket_process_ctrl_pop": ([c_void_p], []),
    "ket_process_ctrl_stack": ([c_void_p], []),
    "ket_process_ctrl_unstack": ([c_void_p], []),
    "ket_process_adj_begin": ([c_void_p], []),
    "ket_process_adj_end": ([c_void_p], []),
    "ket_process_execute": ([c_void_p], []),
    "ket_process_transpile": ([c_void_p], []),
    "ket_process_instructions_json": (
        [c_void_p, POINTER(c_uint8), c_size_t],
        [c_size_t],
    ),
    "ket_process_isa_instructions_json": (
        [c_void_p, POINTER(c_uint8), c_size_t],
        [c_size_t],
    ),
    "ket_process_metadata_json": ([c_void_p, POINTER(c_uint8), c_size_t], [c_size_t]),
    "ket_process_get_qubit_status": ([c_void_p, c_size_t], [c_bool, c_bool]),
    "ket_process_get_measurement": ([c_void_p, c_size_t], [c_bool, c_uint64]),
    "ket_process_get_exp_value": ([c_void_p, c_size_t], [c_bool, c_double]),
    "ket_process_get_sample": (
        [c_void_p, c_size_t],
        [c_bool, POINTER(c_uint64), POINTER(c_uint64), c_size_t],
    ),
    "ket_process_get_dump_size": ([c_void_p, c_size_t], [c_bool, c_size_t]),
    "ket_process_get_dump": (
        [c_void_p, c_size_t, c_size_t],
        [POINTER(c_uint64), c_size_t, c_double, c_double],
    ),
    "ket_make_configuration": (
        [
            c_size_t,  # num_qubits
            POINTER(BatchCExecution),  # batch_execution
            c_int32,  # measure
            c_int32,  # sample
            c_int32,  # exp_value
            c_int32,  # dump
            c_int32,  # gradient
            c_bool,  # define_qpu
            POINTER(c_size_t),  # coupling_graph
            c_size_t,  # coupling_graph_size
            c_int32,  # u4_gate
            c_int32,  # u2_gates
        ],
        [c_void_p],
    ),
}


def libket_path():
    """Get Libket shared library path"""

    if "LIBKET_PATH" in environ:
        path = environ["LIBKET_PATH"]
    else:
        path = f'{dirname(__file__)}/libs/{os_lib_name("ket")}'

    return path


API = load_lib("Libket", libket_path(), API_argtypes, "ket_error_message")


def set_log(level: int):
    """Set Libket log level"""

    API["ket_set_log_level"](level)


class Process:
    """Libket process wrapper from C API"""

    def __init__(self, configuration):
        self._as_parameter_ = API["ket_process_new"](configuration)
        self._finalizer = weakref.finalize(
            self, API["ket_process_delete"], self._as_parameter_
        )

    def __getattr__(self, name: str):
        return lambda *args: API["ket_process_" + name](self, *args)

    def __repr__(self) -> str:
        return "<Libket 'process'>"


class BatchExecution(ABC):
    """Base class for constructing batch target executions."""

    def __init__(self):
        @CFUNCTYPE(
            None,
            POINTER(c_uint8),
            c_size_t,
            POINTER(c_uint8),
            c_size_t,
            POINTER(c_double),
            c_size_t,
        )
        def submit_execution(  # pylint: disable=too-many-arguments,too-many-positional-arguments
            logical_circuit,
            logical_circuit_size,
            physical_circuit,
            physical_circuit_size,
            parameters,
            parameters_size,
        ):
            logical_circuit = json.loads(
                bytearray(logical_circuit[:logical_circuit_size])
            )
            physical_circuit = json.loads(
                bytearray(physical_circuit[:physical_circuit_size])
            )
            parameters = parameters[:parameters_size]
            self.submit_execution(logical_circuit, physical_circuit, parameters)

        self._result_json = None
        self._result_len = None

        @CFUNCTYPE(None, POINTER(POINTER(c_uint8)), POINTER(c_size_t))
        def get_result(result_ptr, size):
            result = self.get_result()
            self._result_json = json.dumps(result).encode("utf-8")
            self._result_len = len(self._result_json)
            self._result_json = (c_uint8 * self._result_len)(*self._result_json)
            result_ptr[0] = self._result_json
            size[0] = self._result_len

        @CFUNCTYPE(None)
        def clear():
            self.clear()

        self.c_struct = BatchCExecution(
            submit_execution,
            get_result,
            clear,
        )

    @abstractmethod
    def submit_execution(
        self,
        logical_circuit: dict,
        physical_circuit: dict | None,
        parameters: list[float],
    ):
        """Get the quantum circuit to execute."""

    @abstractmethod
    def get_result(self) -> dict:
        """Get the result of the quantum circuit execution."""

    @abstractmethod
    def clear(self):
        """Clear the data to start a new execution."""

    def configure(self, **kwargs):
        """Configure the batch execution."""
        return make_configuration(batch_execution=self.c_struct, **kwargs)

    @staticmethod
    def get_qubit_index(qubit):
        """Get the qubit index from the qubit object."""
        match qubit:
            case {"Main": {"index": index}}:
                ...
            case {"index": index}:
                ...
            case _:
                raise ValueError(f"Invalid qubit: {qubit}")
        return index

    @staticmethod
    def get_gate_and_angle(gate):
        """Get the gate and possible angle from the gate object."""
        if isinstance(gate, str):
            return gate, {}
        return list(gate.items())[0]

    def process_instructions(self, instructions: dict):
        """Parse the instructions to execute."""
        for instruction in instructions:
            match instruction:
                case {"Gate": {"control": control, "gate": gate, "target": target}}:
                    control = list(map(self.get_qubit_index, control))
                    target = self.get_qubit_index(target)
                    gate, param = self.get_gate_and_angle(gate)
                    match gate:
                        case "Hadamard":
                            self.hadamard(target, control, **param)
                        case "PauliX":
                            self.pauli_x(target, control, **param)
                        case "PauliY":
                            self.pauli_y(target, control, **param)
                        case "PauliZ":
                            self.pauli_z(target, control, **param)
                        case "RotationX":
                            self.rotation_x(target, control, **param)
                        case "RotationY":
                            self.rotation_y(target, control, **param)
                        case "RotationZ":
                            self.rotation_z(target, control, **param)
                        case "Phase":
                            self.phase(target, control, **param)
                case {"ExpValue": {"index": index, "hamiltonian": hamiltonian}}:
                    self.exp_value(index, hamiltonian)
                case {"Measure": {"index": index, "qubits": qubits}}:
                    qubits = list(map(self.get_qubit_index, qubits))
                    self.measure(index, qubits)
                case {"Sample": {"index": index, "qubits": qubits, "shots": shots}}:
                    qubits = list(map(self.get_qubit_index, qubits))
                    self.sample(index, qubits, shots)
                case {"Dump": {"index": index, "qubits": qubits}}:
                    qubits = list(map(self.get_qubit_index, qubits))
                    self.dump(index, qubits)
                case "Identity":
                    continue
                case _:
                    raise ValueError(f"Invalid instruction: {instruction}")

    def pauli_x(self, target, control):
        """Apply a Pauli-X gate to the target qubit."""
        raise NotImplementedError("Pauli-X gate not implemented")

    def pauli_y(self, target, control):
        """Apply a Pauli-Y gate to the target qubit."""
        raise NotImplementedError("Pauli-Y gate not implemented")

    def pauli_z(self, target, control):
        """Apply a Pauli-Z gate to the target qubit."""
        raise NotImplementedError("Pauli-Z gate not implemented")

    def hadamard(self, target, control):
        """Apply a Pauli-Z gate to the target qubit."""
        raise NotImplementedError("Hadamard gate not implemented")

    def rotation_x(self, target, control, **kwargs):
        """Apply a X-Rotation gate to the target qubit."""
        raise NotImplementedError("X-Rotation gate not implemented")

    def rotation_y(self, target, control, **kwargs):
        """Apply a Y-Rotation gate to the target qubit."""
        raise NotImplementedError("Y-Rotation gate not implemented")

    def rotation_z(self, target, control, **kwargs):
        """Apply a Z-Rotation gate to the target qubit."""
        raise NotImplementedError("Z-Rotation gate not implemented")

    def phase(self, target, control, **kwargs):
        """Apply a Phase gate to the target qubit."""
        raise NotImplementedError("Phase gate not implemented")

    def exp_value(self, index, hamiltonian):
        """Compute the expectation value."""
        raise NotImplementedError("Expectation value not implemented")

    def measure(self, index, qubits) -> int:
        """Measure the qubits."""
        raise NotImplementedError("Measurement not implemented")

    def sample(self, index, qubits, shots):
        """Sample the qubits."""
        raise NotImplementedError("Sampling not implemented")

    def dump(self, index, qubits):
        """Dump the qubits."""
        raise NotImplementedError("Dump not implemented")


_FEATURE_STATUS = {"Disable": 0, "Allowed": 1, "ValidAfter": 2}


def make_configuration(  # pylint: disable=too-many-arguments,too-many-positional-arguments
    num_qubits: int,
    batch_execution,
    measure: Literal["Disable", "Allowed", "ValidAfter"],
    sample: Literal["Disable", "Allowed", "ValidAfter"],
    exp_value: Literal["Disable", "Allowed", "ValidAfter"],
    dump: Literal["Disable", "Allowed", "ValidAfter"],
    gradient: Literal["Disable", "ParameterShift", "SupportsGradient"],
    define_qpu: bool,
    coupling_graph: list[tuple[int, int]] | None,
    u4_gate_type: Literal["CX", "CZ"],
    u2_gate_set: Literal["All", "ZYZ", "RzSx"],
) -> Process:
    """Make a Libket configuration"""

    coupling_graph_size = len(coupling_graph) if coupling_graph else 0
    if coupling_graph_size > 0:
        coupling_graph = reduce(iconcat, coupling_graph, [])
        coupling_graph = (c_size_t * len(coupling_graph))(*coupling_graph)
    else:
        coupling_graph = None

    return API["ket_make_configuration"](
        num_qubits,  # num_qubits
        batch_execution,  # batch_execution
        _FEATURE_STATUS[measure],  # measure
        _FEATURE_STATUS[sample],  # sample
        _FEATURE_STATUS[exp_value],  # exp_value
        _FEATURE_STATUS[dump],  # dump
        {"Disable": 0, "ParameterShift": 1, "SupportsGradient": 2}[
            gradient
        ],  # gradient
        define_qpu,  # define_qpu
        coupling_graph,  # coupling_graph
        coupling_graph_size,  # coupling_graph_size
        1 if u4_gate_type == "CZ" else 0,  # u4_gate
        {"All": 0, "ZYZ": 1, "RzSx": 2}[u2_gate_set],  # u2_gates
    )
